#!/usr/bin/env python
#-*- coding:utf-8 -*-

import re
import requests
from libnum import len_in_bits, s2n, n2s
from gcm import gcm_decrypt
from oaep import oaep_decode, hash_len
from blocktype_attack import BlocktypeAttack

url = "http://cryptochallenge.nds.rub.de:50080/service"

# {"alg":"RSA_OAEP","iv":"yv66vvrO263eyviI","typ":"JWT","enc":"A128GCM"}
data = """eyJhbGciOiJSU0FfT0FFUCIsIml2IjoieXY2NnZ2ck8yNjNleXZpSSIsInR5cCI6IkpXVCIsImVuYyI6IkExMjhHQ00ifQ==.
ZBnPlwONWHxGDrtCxxopS4y4SrMZIAhUg3HI+SbLMxfPVRPW8yunejrkmfSLO1H/0tOx4ssggygHjG7sUfxL8A==.
i2vygn2vqFpsmep3etrD5Yh5xLP9xYhJdvn63WmHEPYChA==."""

n = 0x8fed3203078bba9fd9a8046da63205afde44a238e03b036c0f1d601415ec3c88c0e9fa82e4f1294c44b03f96a1a51f88a03ef9d36d840658a0a932951ba81081
e = 0x10001
key_size = len_in_bits(n) / 8

def split(data):
    s = [x.rstrip(".").decode("base-64") for x in data.split()]
    return s

def join(stuff, rsa, aes):
    s = ""
    for x in stuff, rsa, aes:
        s += x.encode("base-64") + ".\n"
    return s[:-1]

stuff, rsa, aes = split(data)

rsa_c = s2n(rsa)

def check_pad(rsa_c):
    stuff = '{"alg":"RSA1_5","iv":"yv66vvrO263eyviI","typ":"JWT","enc":"A128GCM"}"'
    r = requests.post(url, data=join(stuff, n2s(rsa_c), aes))
    s = r.content

    if "Blocktype mismatch" in s:
        d = re.findall(r"Blocktype mismatch: (\-?\d+)", s)[0]
        return int(d) % 256
    elif "symmetric key length mismatch" in s or "Padding string not terminated" in s or "Padding string too short" in s:
        return 2
    elif "Data must start with zero" in s:
        return False
    else:
        print "Unknown error for:", rsa_c
        print s
        quit()

b = BlocktypeAttack(n, e, check_pad)
m = b.decrypt(rsa_c)

print "Bleichenbacher attack finished with", b.oracle_requests_count, "oracle requests"
assert (pow(m, e, n) == rsa_c)

data, seed = oaep_decode(n2s(m).rjust(64, "\x00"), hash_len)
key = data[-16:]

iv = "yv66vvrO263eyviI".decode("base-64")
ct = "8b6bf2827dafa85a6c99ea777adac3e58879c4b3fdc5884976f9fadd698710f60284".decode("hex")
ct, tag = ct[:18], ct[18:]
print "PLAINTEXT:", gcm_decrypt(key, iv, ct, "", tag)